import {parser} from "./parser.js"
import {foldNodeProp, foldInside, indentNodeProp, LRLanguage, LanguageSupport, HighlightStyle} from "@codemirror/language"
import {completeFromList, ifIn, ifNotIn} from '@codemirror/autocomplete'
import { NodeProp } from "@lezer/common"
import {styleTags, tags, Tag} from "@lezer/highlight"

export const edgeSpecifier = Tag.define(tags.modifier);
export const portSpecifier = Tag.define(tags.modifier);
export const netTypeName = Tag.define(tags.typeName);
export const netName = Tag.define(tags.name);
export const taskName = Tag.define(tags.name);
export const primitiveType = Tag.define(tags.typeName);
export const moduleName = Tag.define(tags.name);
export const attribute = Tag.define(tags.annotation);

const tagAssociations = {
    Keyword: tags.keyword,
    Identifier: netName,
    VariableIdentifier: tags.variableName,
    DirectiveName: tags.processingInstruction,
    'ParamAssignment/ParameterIdentifier': tags.definition(tags.constant(tags.name)),
    'localparam parameter': tags.definitionKeyword,
    SystemTaskIdentifier: tags.standard(taskName),
    SystemFunctionIdentifier: tags.standard(tags.function(tags.name)),
    UnsignedNumber: tags.number,
    RealNumber: tags.float,
    DecimalValue: tags.integer,
    BinaryValue: tags.integer,
    OctalValue: tags.integer,
    HexValue: tags.integer,
    BlockComment: tags.blockComment,
    LineComment: tags.lineComment,
    DocComment: tags.docComment,
    'always initial': tags.definitionKeyword,
    'edge posedge negedge': edgeSpecifier,
    'cmos rcmos': tags.standard(primitiveType),
    'bufif0 bufif1 notif0 notif1': tags.standard(primitiveType),
    'nmos pmos rnmos rpmos': tags.standard(primitiveType),
    'and nand or nor xor xnor': tags.standard(primitiveType),
    'UDPIdentifier': primitiveType,
    'UDPDeclaration/UDPIdentifier': tags.definition(primitiveType),
    'ModuleIdentifier': moduleName,
    'ModuleDeclaration/ModuleIdentifier': tags.definition(moduleName),
    'ModuleInstanceIdentifier UDPInstanceIdentifier': tags.definition(tags.name),
    '@': tags.special(tags.operator),
    String: tags.string,
    'always initial if case casex casez else': tags.controlKeyword,
    'supply0 supply1 tri triand trior tri0 tri1 uwire wire wand wor reg': netTypeName,
    '( )': tags.paren,
    '(* *)': tags.paren,
    'module macromodule endmodule': tags.moduleKeyword,
    'input output inout': portSpecifier,
    'x X z Z': tags.number,
    AttrSpec: attribute,
    'VariableIdentifier OutputPortIdentifier': tags.variableName
}

/**
 * 
 * @param {string[]} pairs 
 * @returns 
 */
const closedBy = (pairs) => 
    NodeProp.closedBy.add(
        Object.fromEntries(
            pairs.map((pair)=>{
                const [selector, ...closers] = pair.split(/\w+/);
                return [selector.replace(',', ' '), closers];
            })
        ))
const augmentedParser = parser.configure({
    props: [
        styleTags(tagAssociations),
        closedBy([
            '( )',
            '[ ]',
            '{ }',
            '" "',
            '/* */',
            '/** */',
            '(* *)',
            'begin end',
            'case,casex,casez endcase',
            'config endconfig',
            'fork join',
            'function endfunction',
            'generate endgenerate',
            'module,macromodule endmodule',
            'primitive endprimitive',
            'specify endspecify',
            'table endtable',
            'task endtask',
        ]),
        foldNodeProp.add({
            [
                'ModuleDeclaration'
                + ' ConfigDeclaration'
                + ' FunctionDeclaration'
                + ' TaskDeclaration'
                + ' UDPDeclaration'
            ]: context => {
                const sc = context.getChild(';');
                const lc = context.lastChild;
                if (!sc || !lc)
                    return null;

                return { from: sc.to, to: lc.from - 1};
            },
            [
                'GenerateRegion'
                + ' CombinationalBody'
                + ' SequentialBody'
                + ' SpecifyBlock'
            ]: context => {
                const fc = context.firstChild;
                const lc = context.lastChild;
                if (!fc || !lc)
                    return null;

                return { from: fc.to, to: lc.from };
            },
            'CaseGenerateConstruct CaseStatement': context => {
                const fc = context.getChild(')');
                const lc = context.lastChild;
                if (!fc || !lc)
                    return null;

                return { from: fc.to, to: lc.from };
            },
            'GenerateBlock': context => {
                const fc = context.getChild('GenerateBlockIdentifier')
                        ?? context.firstChild;
                const lc = context.lastChild;
                if (!fc || !lc)
                    return null;

                return { from: fc.to, to: lc.from };
            },
            'ParBlock SeqBlock': context => {
                const fc = context.getChild('BlockIdentifier') ?? context.firstChild;
                const lc = context.lastChild;
                if (!fc || !lc)
                    return null;

                return { from: fc.to, to: lc.from };
            }
        }),
        indentNodeProp.add({
            ['ModuleDeclaration'
                + ' GenerateRegion'
                + ' TaskDeclaration'
                + ' FunctionDeclaration'
                + ' ConfigDeclaration'
                + ' UDPDeclaration'
                + ' CombinationalBody'
                + ' SequentialBody'
                + ' SpecifyBlock'
                + ' AttributeInstance'
                + ' ParBlock'
                + ' SeqBlock'
                + ' CaseStatement'
                + ' CaseGenerateConstruct'
                + ' ParameterDeclaration'
                + ' SpecParamDeclaration'
                + ' UDPInstantiation'
                + ' ModuleInstantiation'
                + ' ConditionalGenerateConstruct'
                + ' Declaration'
            ]: 
                context => (context.continue() ?? 0) + context.unit
        })
    ]
})

export const verilogLanguage = LRLanguage.define({
    parser: augmentedParser
})

export const verilogCompletion = verilogLanguage.data.of({
    autocomplete: [
        completeFromList([
        { label: 'begin', type: 'keyword' },
        { label: 'module', type: 'keyword' },
        { label: 'end', type: 'keyword' },
        { label: 'endmodule', type: 'keyword' }
    ])
]
})

export const verilogHighlightStyle = HighlightStyle.define([
    { tag: tags.name, class: 'cm-tag-name' },
    { tag: tags.keyword, class: 'cm-tag-keyword' },
    { tag: tags.blockComment, class: 'cm-tag-block-comment' },
    { tag: tags.docComment, class: 'cm-tag-doc-comment' },
    { tag: tags.lineComment, class: 'cm-tag-line-comment' },
    { tag: tags.annotation, class: 'cm-tag-annotation'}
])

export function verilog() {
    return new LanguageSupport(verilogLanguage, [verilogCompletion]);
}